@@ -88,12 +88,18 @@ void stmt_set_error(MYSQL_STMT *stmt,
                   ...)
 {
   va_list ap;
+  const char *error= NULL;
+
+  if (error_nr >= CR_MIN_ERROR && error_nr <= CR_MYSQL_LAST_ERROR)
+    error= ER(error_nr);
+  else if (error_nr >= CER_MIN_ERROR && error_nr <= CR_MARIADB_LAST_ERROR)
+    error= CER(error_nr);
 
   stmt->last_errno= error_nr;
   ma_strmake(stmt->sqlstate, sqlstate, SQLSTATE_LENGTH);
   va_start(ap, format);
   vsnprintf(stmt->last_error, MYSQL_ERRMSG_SIZE,
-            format ? format : ER(error_nr), ap);
+            format ? format : error ? error : "", ap);
   va_end(ap);
   return;
 }
@@ -246,7 +252,7 @@ int mthd_stmt_read_all_rows(MYSQL_STMT *stmt)
               {
                 size_t len= MAX(stmt->fields[i].length, mysql_ps_fetch_functions[stmt->fields[i].type].max_len);
                 if (len > stmt->fields[i].max_length)
-                  stmt->fields[i].max_length= len;
+                  stmt->fields[i].max_length= (unsigned long)len;
               }
               else if (!stmt->fields[i].max_length)
               {
@@ -340,10 +346,15 @@ void mthd_stmt_flush_unbuffered(MYSQL_STMT *stmt)
     }
     if (packet_len < 8 && *pos == 254) /* EOF */
     {
-      stmt->mysql->server_status= uint2korr(pos + 3);
-      if (in_resultset)
+      if (mariadb_connection(stmt->mysql))
+      {
+        stmt->mysql->server_status= uint2korr(pos + 3);
+        if (in_resultset)
+          goto end;
+        in_resultset= 1;
+      }
+      else
         goto end;
-      in_resultset= 1;
     }
   }
 end:
@@ -722,7 +733,7 @@ unsigned char* mysql_stmt_execute_generate_simple_request(MYSQL_STMT *stmt, size
       }
       for (i = 0; i < stmt->param_count; i++)
       {
-        /* this differs from mysqlnd, c api supports unsinged !! */
+        /* this differs from mysqlnd, c api supports unsigned !! */
         uint buffer_type= stmt->params[i].buffer_type | (stmt->params[i].is_unsigned ? 32768 : 0);
         /* check if parameter requires indicator variable */
         int2store(p, buffer_type);
@@ -864,7 +875,7 @@ unsigned char* mysql_stmt_execute_generate_bulk_request(MYSQL_STMT *stmt, size_t
   if (!stmt->param_count)
   {
     stmt_set_error(stmt, CR_BULK_WITHOUT_PARAMETERS, "IM001",
-                   CER(CR_BULK_WITHOUT_PARAMETERS), "Bulk operation");
+                   CER(CR_BULK_WITHOUT_PARAMETERS));
     return NULL;
   }
 
@@ -907,7 +918,7 @@ unsigned char* mysql_stmt_execute_generate_bulk_request(MYSQL_STMT *stmt, size_t
       }
       for (i = 0; i < stmt->param_count; i++)
       {
-        /* this differs from mysqlnd, c api supports unsinged !! */
+        /* this differs from mysqlnd, c api supports unsigned !! */
         uint buffer_type= stmt->params[i].buffer_type | (stmt->params[i].is_unsigned ? 32768 : 0);
         int2store(p, buffer_type);
         p+= 2;
@@ -1323,15 +1334,18 @@ static my_bool net_stmt_close(MYSQL_STMT *stmt, my_bool remove)
 
 my_bool STDCALL mysql_stmt_close(MYSQL_STMT *stmt)
 {
-  my_bool rc;
-  if (stmt && stmt->mysql && stmt->mysql->net.pvio)
-    mysql_stmt_internal_reset(stmt, 1);
+  my_bool rc= 1;
 
-  rc= net_stmt_close(stmt, 1);
+  if (stmt)
+  {
+    if (stmt->mysql && stmt->mysql->net.pvio)
+      mysql_stmt_internal_reset(stmt, 1);
 
-  free(stmt->extension);
-  free(stmt);
+    rc= net_stmt_close(stmt, 1);
 
+    free(stmt->extension);
+    free(stmt);
+  }
   return(rc);
 }
 
@@ -1395,12 +1409,15 @@ int STDCALL mysql_stmt_fetch(MYSQL_STMT *stmt)
     return(rc);
   }
 
-  rc= stmt->mysql->methods->db_stmt_fetch_to_bind(stmt, row);
+  if ((rc= stmt->mysql->methods->db_stmt_fetch_to_bind(stmt, row)))
+  {
+    return(rc);
+  }
 
   stmt->state= MYSQL_STMT_USER_FETCHING;
   CLEAR_CLIENT_ERROR(stmt->mysql);
   CLEAR_CLIENT_STMT_ERROR(stmt);
-  return(rc);
+  return(0);
 }
 
 int STDCALL mysql_stmt_fetch_column(MYSQL_STMT *stmt, MYSQL_BIND *bind, unsigned int column, unsigned long offset)
@@ -1496,7 +1513,7 @@ my_bool mthd_stmt_read_prepare_response(MYSQL_STMT *stmt)
 
   p= (uchar *)stmt->mysql->net.read_pos;
 
-  if (0xFF == p[0])  /* Error occured */
+  if (0xFF == p[0])  /* Error occurred */
   {
     return(1);
   }
@@ -1573,7 +1590,7 @@ int STDCALL mysql_stmt_prepare(MYSQL_STMT *stmt, const char *query, unsigned lon
     is_multi= (mysql->net.extension->multi_status > COM_MULTI_OFF);
     /* We need to semi-close the prepared statement:
        reset stmt and free all buffers and close the statement
-       on server side. Statment handle will get a new stmt_id */
+       on server side. Statement handle will get a new stmt_id */
 
     if (!is_multi)
       ma_multi_command(mysql, COM_MULTI_ENABLED);
@@ -1797,11 +1814,11 @@ int stmt_read_execute_response(MYSQL_STMT *stmt)
 
   ret= test((mysql->methods->db_read_stmt_result &&
                  mysql->methods->db_read_stmt_result(mysql)));
-  /* if a reconnect occured, our connection handle is invalid */
+  /* if a reconnect occurred, our connection handle is invalid */
   if (!stmt->mysql)
     return(1);
 
-  /* update affected rows, also if an error occured */
+  /* update affected rows, also if an error occurred */
   stmt->upsert_status.affected_rows= stmt->mysql->affected_rows;
 
   if (ret)
@@ -1846,6 +1863,12 @@ int stmt_read_execute_response(MYSQL_STMT *stmt)
 
       for (i=0; i < stmt->field_count; i++)
       {
+        memcpy(&stmt->fields[i], &mysql->fields[i], sizeof(MYSQL_FIELD));
+
+        /* since  all pointers will be incorrect if another statement will
+           be executed, so we need to allocate memory and copy the
+           information */
+        stmt->fields[i].extension= 0; /* not in use yet */
         if (mysql->fields[i].db)
           stmt->fields[i].db= ma_strdup_root(fields_ma_alloc_root, mysql->fields[i].db);
         if (mysql->fields[i].table)
@@ -1858,7 +1881,8 @@ int stmt_read_execute_response(MYSQL_STMT *stmt)
           stmt->fields[i].org_name= ma_strdup_root(fields_ma_alloc_root, mysql->fields[i].org_name);
         if (mysql->fields[i].catalog)
           stmt->fields[i].catalog= ma_strdup_root(fields_ma_alloc_root, mysql->fields[i].catalog);
-        stmt->fields[i].def= mysql->fields[i].def ? ma_strdup_root(fields_ma_alloc_root, mysql->fields[i].def) : NULL;
+        if (mysql->fields[i].def)
+          stmt->fields[i].def= ma_strdup_root(fields_ma_alloc_root, mysql->fields[i].def);
       }
     }
 
@@ -1883,7 +1907,8 @@ int stmt_read_execute_response(MYSQL_STMT *stmt)
          */
 
       /* preferred is buffered read */
-      mysql_stmt_store_result(stmt);
+      if (mysql_stmt_store_result(stmt))
+        return 1;
       stmt->mysql->status= MYSQL_STATUS_STMT_RESULT;
     } else
     {
@@ -1962,8 +1987,9 @@ int STDCALL mysql_stmt_execute(MYSQL_STMT *stmt)
   {
     ma_free_root(&stmt->result.alloc, MYF(MY_KEEP_PREALLOC));
     stmt->result_cursor= stmt->result.data= 0;
-    stmt->result.rows= 0;
   }
+  /* CONC-344: set row count to zero */
+  stmt->result.rows= 0;
   if (stmt->array_size > 0)
     request= (char *)mysql_stmt_execute_generate_bulk_request(stmt, &request_len);
   else
@@ -2276,7 +2302,7 @@ int STDCALL mysql_stmt_next_result(MYSQL_STMT *stmt)
   }
 
   if (stmt->mysql->status == MYSQL_STATUS_GET_RESULT)
-    stmt->mysql->status= MYSQL_STATUS_STMT_RESULT;
+    stmt->mysql->status= MYSQL_STATUS_STMT_RESULT; 
 
   if (stmt->mysql->field_count)
     rc= madb_alloc_stmt_fields(stmt);
@@ -2297,17 +2323,23 @@ int STDCALL mariadb_stmt_execute_direct(MYSQL_STMT *stmt,
                                       const char *stmt_str,
                                       size_t length)
 {
-  MYSQL *mysql= stmt->mysql;
-  my_bool emulate_cmd= !(!(stmt->mysql->server_capabilities & CLIENT_MYSQL) &&
-      (stmt->mysql->extension->mariadb_server_capabilities &
-      (MARIADB_CLIENT_STMT_BULK_OPERATIONS >> 32)));
+  MYSQL *mysql;
+  my_bool emulate_cmd;
 
+  if (!stmt)
+    return 1;
+
+  mysql= stmt->mysql;
   if (!mysql)
   {
     SET_CLIENT_STMT_ERROR(stmt, CR_SERVER_LOST, SQLSTATE_UNKNOWN, 0);
-    goto fail;
+    return 1;
   }
 
+  emulate_cmd= !(!(stmt->mysql->server_capabilities & CLIENT_MYSQL) &&
+      (stmt->mysql->extension->mariadb_server_capabilities &
+      (MARIADB_CLIENT_STMT_BULK_OPERATIONS >> 32))) || mysql->net.compress;
+
   /* Server versions < 10.2 don't support execute_direct, so we need to 
      emulate it */
   if (emulate_cmd)
@@ -2315,7 +2347,6 @@ int STDCALL mariadb_stmt_execute_direct(MYSQL_STMT *stmt,
     int rc;
 
     /* avoid sending close + prepare in 2 packets */
-
     if ((rc= mysql_stmt_prepare(stmt, stmt_str, (unsigned long)length)))
       return rc;
     return mysql_stmt_execute(stmt);
@@ -2324,13 +2355,7 @@ int STDCALL mariadb_stmt_execute_direct(MYSQL_STMT *stmt,
   if (ma_multi_command(mysql, COM_MULTI_ENABLED))
   {
     SET_CLIENT_STMT_ERROR(stmt, CR_COMMANDS_OUT_OF_SYNC, SQLSTATE_UNKNOWN, 0);
-    goto fail;
-  }
-
-  if (!stmt->mysql)
-  {
-    SET_CLIENT_STMT_ERROR(stmt, CR_SERVER_LOST, SQLSTATE_UNKNOWN, 0);
-    return(1);
+    return 1;
   }
 
   if (length == (size_t) -1)
@@ -2346,7 +2371,7 @@ int STDCALL mariadb_stmt_execute_direct(MYSQL_STMT *stmt,
   {
     /* We need to semi-close the prepared statement:
        reset stmt and free all buffers and close the statement
-       on server side. Statment handle will get a new stmt_id */
+       on server side. Statement handle will get a new stmt_id */
     char stmt_id[STMT_ID_LENGTH];
 
     if (mysql_stmt_internal_reset(stmt, 1))
@@ -2414,8 +2439,10 @@ int STDCALL mariadb_stmt_execute_direct(MYSQL_STMT *stmt,
   /* read execute response packet */
   return stmt_read_execute_response(stmt);
 fail:
-  SET_CLIENT_STMT_ERROR(stmt, mysql->net.last_errno, mysql->net.sqlstate,
-      mysql->net.last_error);
+  /* check if we need to set error message */
+  if (!mysql_stmt_errno(stmt))
+    SET_CLIENT_STMT_ERROR(stmt, mysql->net.last_errno, mysql->net.sqlstate,
+        mysql->net.last_error);
   do {
     stmt->mysql->methods->db_stmt_flush_unbuffered(stmt);
   } while(mysql_stmt_more_results(stmt));
