@@ -42,7 +42,7 @@
 #ifdef HAVE_TLS_SESSION_CACHE
 #undef HAVE_TLS_SESSION_CACHE
 #endif
-#if OPENSSL_USE_BIOMETHOD
+#if defined(OPENSSL_USE_BIOMETHOD)
 #undef OPENSSL_USE_BIOMETHOD
 #endif
 #ifndef HAVE_OPENSSL_DEFAULT
@@ -56,6 +56,9 @@
 #endif
 #include <ma_pthread.h>
 
+#include <mariadb_async.h>
+#include <ma_context.h>
+
 extern my_bool ma_tls_initialized;
 extern unsigned int mariadb_deinitialize_ssl;
 
@@ -66,7 +69,7 @@ static pthread_mutex_t LOCK_openssl_config;
 #ifndef HAVE_OPENSSL_1_1_API
 static pthread_mutex_t *LOCK_crypto= NULL;
 #endif
-#if OPENSSL_USE_BIOMETHOD
+#if defined(OPENSSL_USE_BIOMETHOD)
 static int ma_bio_read(BIO *h, char *buf, int size);
 static int ma_bio_write(BIO *h, const char *buf, int size);
 static BIO_METHOD ma_BIO_method;
@@ -79,7 +82,15 @@ static long ma_tls_version_options(const char *version)
        disable_all_protocols;
 
   protocol_options= disable_all_protocols=
-    SSL_OP_NO_TLSv1 | SSL_OP_NO_TLSv1_1 | SSL_OP_NO_TLSv1_2;
+    SSL_OP_NO_SSLv2 |
+    SSL_OP_NO_SSLv3 |
+    SSL_OP_NO_TLSv1 |
+    SSL_OP_NO_TLSv1_1 |
+    SSL_OP_NO_TLSv1_2
+#ifdef TLS1_3_VERSION
+    | SSL_OP_NO_TLSv1_3
+#endif
+    ;
 
   if (!version)
     return 0;
@@ -90,6 +101,10 @@ static long ma_tls_version_options(const char *version)
     protocol_options&= ~SSL_OP_NO_TLSv1_1;
   if (strstr(version, "TLSv1.2"))
     protocol_options&= ~SSL_OP_NO_TLSv1_2;
+#ifdef TLS1_3_VERSION
+  if (strstr(version, "TLSv1.3"))
+    protocol_options&= ~SSL_OP_NO_TLSv1_3;
+#endif
 
   if (protocol_options != disable_all_protocols)
     return protocol_options;
@@ -255,7 +270,11 @@ static void my_cb_locking(int mode, int n,
 
 static int ssl_thread_init()
 {
-  if (!CRYPTO_get_id_callback())
+  if (!CRYPTO_THREADID_get_callback()
+#ifndef OPENSSL_NO_DEPRECATED
+      && !CRYPTO_get_id_callback()
+#endif
+      )
   {
     int i, max= CRYPTO_num_locks();
 
@@ -339,7 +358,7 @@ int ma_tls_start(char *errmsg __attribute__((unused)), size_t errmsg_len __attri
   OpenSSL_add_all_algorithms();
 #endif
   disable_sigpipe();
-#if OPENSSL_USE_BIOMETHOD
+#ifdef OPENSSL_USE_BIOMETHOD
   memcpy(&ma_BIO_method, BIO_s_socket(), sizeof(BIO_METHOD));
   ma_BIO_method.bread= ma_bio_read;
   ma_BIO_method.bwrite= ma_bio_write;
@@ -382,7 +401,8 @@ void ma_tls_end()
     {
       int i;
       CRYPTO_set_locking_callback(NULL);
-      CRYPTO_set_id_callback(NULL);
+      CRYPTO_THREADID_set_callback(NULL);
+
       for (i=0; i < CRYPTO_num_locks(); i++)
         pthread_mutex_destroy(&LOCK_crypto[i]);
       ma_free((gptr)LOCK_crypto);
@@ -431,9 +451,15 @@ static int ma_tls_set_certs(MYSQL *mysql, SSL *ssl)
   
   /* add cipher */
   if ((mysql->options.ssl_cipher &&
-        mysql->options.ssl_cipher[0] != 0) &&
+        mysql->options.ssl_cipher[0] != 0))
+   {
+     if(
+#ifdef TLS1_3_VERSION
+      SSL_set_ciphersuites(ssl, mysql->options.ssl_cipher) == 0 &&
+#endif
       SSL_set_cipher_list(ssl, mysql->options.ssl_cipher) == 0)
     goto error;
+   }
 
   /* ca_file and ca_path */
   if (!SSL_CTX_load_verify_locations(ctx,
@@ -512,7 +538,9 @@ void *ma_tls_init(MYSQL *mysql)
 {
   SSL *ssl= NULL;
   SSL_CTX *ctx= NULL;
-  long options= SSL_OP_ALL;
+  long options= SSL_OP_ALL |
+                SSL_OP_NO_SSLv2 |
+                SSL_OP_NO_SSLv3;
 #ifdef HAVE_TLS_SESSION_CACHE
   MA_SSL_SESSION *session= ma_tls_get_session(mysql);
 #endif
@@ -568,7 +596,7 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
   MYSQL *mysql;
   MARIADB_PVIO *pvio;
   int rc;
-#if OPENSSL_USE_BIOMETHOD
+#ifdef OPENSSL_USE_BIOMETHOD
   BIO_METHOD *bio_method= NULL;
   BIO *bio;
 #endif
@@ -582,7 +610,7 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
 
   SSL_clear(ssl);
 
-#if OPENSSL_USE_BIOMETHOD
+#ifdef OPENSSL_USE_BIOMETHOD
   bio= BIO_new(&ma_BIO_method);
   bio->ptr= pvio;
   SSL_set_bio(ssl, bio, bio);
@@ -633,14 +661,91 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
   return 0;
 }
 
+static my_bool
+ma_tls_async_check_result(int res, struct mysql_async_context *b, SSL *ssl)
+{
+  int ssl_err;
+  b->events_to_wait_for= 0;
+  if (res >= 0)
+    return 1;
+  ssl_err= SSL_get_error(ssl, res);
+  if (ssl_err == SSL_ERROR_WANT_READ)
+    b->events_to_wait_for|= MYSQL_WAIT_READ;
+  else if (ssl_err == SSL_ERROR_WANT_WRITE)
+    b->events_to_wait_for|= MYSQL_WAIT_WRITE;
+  else
+    return 1;
+  if (b->suspend_resume_hook)
+    (*b->suspend_resume_hook)(TRUE, b->suspend_resume_hook_user_data);
+  my_context_yield(&b->async_context);
+  if (b->suspend_resume_hook)
+    (*b->suspend_resume_hook)(FALSE, b->suspend_resume_hook_user_data);
+  return 0;
+}
+
+ssize_t ma_tls_read_async(MARIADB_PVIO *pvio,
+                          const unsigned char *buffer,
+                          size_t length)
+{
+  int res;
+  struct mysql_async_context *b= pvio->mysql->options.extension->async_context;
+  MARIADB_TLS *ctls= pvio->ctls;
+
+  for (;;)
+  {
+    res= SSL_read((SSL *)ctls->ssl, (void *)buffer, length);
+    if (ma_tls_async_check_result(res, b, (SSL *)ctls->ssl))
+      return res;
+  }
+}
+
+ssize_t ma_tls_write_async(MARIADB_PVIO *pvio,
+                           const unsigned char *buffer,
+                           size_t length)
+{
+  int res;
+  struct mysql_async_context *b= pvio->mysql->options.extension->async_context;
+  MARIADB_TLS *ctls= pvio->ctls;
+
+  for (;;)
+  {
+    res= SSL_write((SSL *)ctls->ssl, (void *)buffer, length);
+    if (ma_tls_async_check_result(res, b, (SSL *)ctls->ssl))
+      return res;
+  }
+}
+
+
 ssize_t ma_tls_read(MARIADB_TLS *ctls, const uchar* buffer, size_t length)
 {
-  return SSL_read((SSL *)ctls->ssl, (void *)buffer, (int)length);
+  ssize_t rc;
+  MARIADB_PVIO *pvio= ctls->pvio;
+
+  while ((rc= SSL_read((SSL *)ctls->ssl, (void *)buffer, (int)length)) < 0)
+  {
+    int error= SSL_get_error((SSL *)ctls->ssl, rc);
+    if (error != SSL_ERROR_WANT_READ)
+      return rc;
+    if (pvio->methods->wait_io_or_timeout(pvio, TRUE, pvio->mysql->options.read_timeout) < 1)
+      return rc;
+  }
+  return rc;
 }
 
 ssize_t ma_tls_write(MARIADB_TLS *ctls, const uchar* buffer, size_t length)
-{ 
-  return SSL_write((SSL *)ctls->ssl, (void *)buffer, (int)length);
+{
+  ssize_t rc;
+  MARIADB_PVIO *pvio= ctls->pvio;
+
+  while ((rc= SSL_write((SSL *)ctls->ssl, (void *)buffer, (int)length)) <= 0)
+  {
+    int error= SSL_get_error((SSL *)ctls->ssl, rc);
+    if (error != SSL_ERROR_WANT_WRITE)
+      return rc;
+    if (pvio->methods->wait_io_or_timeout(pvio, TRUE, pvio->mysql->options.write_timeout) < 1)
+      return rc;
+  }
+  return rc;
 }
 
 my_bool ma_tls_close(MARIADB_TLS *ctls)
